import logging
from functools import wraps
from io import BytesIO
from pathlib import Path
from threading import Lock as ThreadLock
from typing import (
    Any,
    AsyncIterator,
    Callable,
    ClassVar,
    Dict,
    List,
    NamedTuple,
    Optional,
    TYPE_CHECKING,
    Tuple,
    TypeVar,
    Union,
)

from aiofiles import open as async_open

# noinspection PyUnresolvedReferences
from aiofiles.tempfile import SpooledTemporaryFile
from aiohttp.typedefs import StrOrURL
from aiolimiter import AsyncLimiter
from arkowrapper import ArkoWrapper
from httpx import USE_CLIENT_DEFAULT

# noinspection PyProtectedMember
from httpx._client import UseClientDefault

# noinspection PyProtectedMember
from httpx._types import (
    AuthTypes,
    CookieTypes,
    HeaderTypes,
    ProxiesTypes,
    QueryParamTypes,
    RequestContent,
    RequestData,
    RequestExtensions,
    RequestFiles,
    TimeoutTypes,
    URLTypes,
)
from typing_extensions import ParamSpec

# noinspection PyProtectedMember
from async_pixiv.client._section import SectionType
from async_pixiv.error import (
    LoginError,
    OauthError,
)
from async_pixiv.model.user import User
from async_pixiv.utils.context import set_pixiv_client
from async_pixiv.utils.net import Net
from async_pixiv.utils.overwrite import Response

if TYPE_CHECKING:
    from async_pixiv.client import (
        USER,
        ILLUST,
        NOVEL,
    )

__all__ = ["PixivClient"]

_REDIRECT_URI = "https://app-api.pixiv.net/web/v1/users/auth/pixiv/callback"
_LOGIN_URL = "https://app-api.pixiv.net/web/v1/login"
_LOGIN_VERIFY = "https://accounts.pixiv.net/ajax/login"
_AUTH_TOKEN_URL = "https://oauth.secure.pixiv.net/auth/token"

P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")
logger = logging.getLogger("async_pixiv.client")


# noinspection PyPep8Naming
class PixivClient(Net):
    _instances: ClassVar[List["PixivClient"]] = []

    @classmethod
    def get_client(cls) -> "PixivClient":
        if cls._instances:
            return cls._instances[0]
        else:
            raise ValueError("请先实列化一个 PixivClient ")

    _lock: ThreadLock = ThreadLock()

    class Config(NamedTuple):
        client_id: str
        client_secret: str
        user_agent: str
        accept_language: str

    _config: Config
    _request_headers: Dict[str, Any]

    access_token: Optional[str] = None
    refresh_token: Optional[str] = None
    account: Optional[User] = None
    _sections: Dict[str, SectionType] = {}

    @property
    def USER(self) -> "USER":
        with self._lock:
            if self._sections.get("user", None) is None:
                from async_pixiv.client import USER

                self._sections["user"] = USER(self)
        return self._sections["user"]

    @property
    def ILLUST(self) -> "ILLUST":
        with self._lock:
            if self._sections.get("illust", None) is None:
                from async_pixiv.client import ILLUST

                self._sections["illust"] = ILLUST(self)
        return self._sections["illust"]

    @property
    def NOVEL(self) -> "NOVEL":
        with self._lock:
            if self._sections.get("novel", None) is None:
                from async_pixiv.client import NOVEL

                self._sections["novel"] = NOVEL(self)
        return self._sections["novel"]

    @property
    def is_logged(self) -> bool:
        return self.refresh_token is not None

    def need_logged(self, func: Callable[[P], R]) -> Callable[[P], R]:
        @wraps(func)
        def wrapper():
            if not self.is_logged:
                raise OauthError("Please login.")
            return func

        return wrapper

    def __init__(
        self,
        *,
        max_rate: float = 100,
        rate_time_period: float = 60,
        timeout: float = 10,
        proxies: Optional[ProxiesTypes] = None,
        trust_env: bool = False,
        retry: int = 5,
        retry_sleep: float = 1,
    ):
        super().__init__(
            rate_limiter=AsyncLimiter(max_rate, rate_time_period),
            timeout=timeout,
            proxies=proxies,
            trust_env=trust_env,
            retry=retry,
            retry_sleep=retry_sleep,
        )
        config_path = (Path(__file__) / "../config").resolve()
        with open(config_path, encoding="utf-8") as file:
            self._config = PixivClient.Config(
                *ArkoWrapper(file.readlines()).map(lambda x: x.strip())
            )
        self._request_headers = {
            "App-OS": "ios",
            "App-OS-Version": "12.2",
            "App-Version": "7.6.2",
            "user-agent": self._config.user_agent,
            "Referer": "https://app-api.pixiv.net/",
            "Accept-Language": "zh-CN,zh;q=0.9,zh-Hans;q=0.8,en;q=0.7,zh-Hant;"
            "q=0.6,ja;q=0.5",
        }
        PixivClient._instances.append(self)

    def set_accept_language(self, language: str) -> None:
        self._request_headers.update({"Accept-Language": language})

    async def request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
        cookies: Optional[CookieTypes] = None,
        auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: Union[UseClientDefault, bool] = USE_CLIENT_DEFAULT,
        timeout: Union[UseClientDefault, TimeoutTypes] = USE_CLIENT_DEFAULT,
        extensions: Optional[RequestExtensions] = None,
    ) -> Response:
        headers = headers or {}
        params = params or {}

        headers.update(
            {
                "Authorization": self.access_token
                if self.access_token is None
                else f"Bearer {self.access_token}",
            }
        )

        for key, value in self._request_headers.items():
            if key not in headers:
                headers[key] = value

        return await super().request(
            method,
            str(url),
            content=content,
            data=data,
            files=files,
            json=json,
            params={k: v for k, v in params.items() if v is not None},
            headers={k: v for k, v in headers.items() if v is not None},
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )

    def stream(
        self,
        method: str,
        url: URLTypes,
        *,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
        cookies: Optional[CookieTypes] = None,
        auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: Union[UseClientDefault, bool] = USE_CLIENT_DEFAULT,
        timeout: Union[UseClientDefault, TimeoutTypes] = USE_CLIENT_DEFAULT,
        extensions: Optional[RequestExtensions] = None,
    ) -> AsyncIterator[Response]:
        headers = headers or {}
        params = params or {}

        headers.update(
            {
                "Authorization": self.access_token
                if self.access_token is None
                else f"Bearer {self.access_token}",
                **self._request_headers,
            }
        )

        return super().stream(
            method,
            str(url),
            content=content,
            data=data,
            files=files,
            json=json,
            params={k: v for k, v in params.items() if v is not None},
            headers={k: v for k, v in headers.items() if v is not None},
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )

    async def login_with_pwd(
        self, username: str, password: str, proxy: Optional[str] = None
    ) -> User:
        try:
            from playwright.async_api import async_playwright
        except ImportError:
            raise ImportError("Please install 'playwright'.")

        def oauth_pkce() -> Tuple[str, str]:
            from secrets import token_urlsafe
            from hashlib import sha256
            from base64 import urlsafe_b64encode

            verifier = token_urlsafe(32)
            challenge = (
                urlsafe_b64encode(sha256(verifier.encode("ascii")).digest())
                .rstrip(b"=")
                .decode("ascii")
            )
            return verifier, challenge

        async with async_playwright() as playwright:
            from urllib.parse import (
                parse_qs,
                urlparse,
                urlencode,
            )

            if proxy is not None:
                # noinspection PyTypeChecker
                browser = await playwright.chromium.launch(proxy={"server": proxy})
            else:
                browser = await playwright.chromium.launch()
            context = await browser.new_context()
            api_request_context = context.request
            page = await context.new_page()

            # 访问登录页面
            code_verifier, code_challenge = oauth_pkce()
            await page.goto(
                urlparse(_LOGIN_URL)
                ._replace(
                    query=urlencode(
                        {
                            "code_challenge": code_challenge,
                            "code_challenge_method": "S256",
                            "client": "pixiv-android",
                        }
                    )
                )
                .geturl(),
                timeout=0,
            )

            # 输入用户名与密码
            await page.locator(".degQSE").fill(username)
            await page.locator(".hfoSmp").fill(password)

            # 点击登录按钮
            # noinspection SpellCheckingInspection
            await page.locator("button.sc-bdnxRM:nth-child(5)").click()

            # 验证登录
            # noinspection PyBroadException
            try:
                async with page.expect_response(f"{_LOGIN_VERIFY}*") as future:
                    response = await (await future.value).json()
                raise LoginError(f"登录错误，请检查登录的用户名和密码是否正确：{response['body']['errors']}")
            except Exception as e:
                if not isinstance(e, LoginError):
                    logger.debug("登录成功，正尝试获取 token")

            # 获取code
            async with page.expect_request(f"{_REDIRECT_URI}*") as request:
                url = urlparse((await request.value).url)
            code = parse_qs(url.query)["code"][0]

            # 获取token
            response = await api_request_context.post(
                _AUTH_TOKEN_URL,
                form={
                    "client_id": self._config.client_id,
                    "client_secret": self._config.client_secret,
                    "code": code,
                    "code_verifier": code_verifier,
                    "grant_type": "authorization_code",
                    "include_policy": "true",
                    "redirect_uri": _REDIRECT_URI,
                },
                headers={
                    "Accept-Encoding": "gzip, deflate",
                    "Content-Type": "application/x-www-form-urlencoded",
                    "User-Agent": self._config.user_agent,
                    "Host": "oauth.secure.pixiv.net",
                },
                timeout=0,
            )
            data = await response.json()
            await browser.close()
        self.access_token = data["access_token"]
        self.refresh_token = data["refresh_token"]
        with set_pixiv_client(self):
            self.account = User.parse_obj(data["user"])
        return self.account

    async def login_with_token(self, token: str) -> User:
        response = await self.post(
            _AUTH_TOKEN_URL,
            data={
                "client_id": self._config.client_id,
                "client_secret": self._config.client_secret,
                "grant_type": "refresh_token",
                "include_policy": "true",
                "refresh_token": token,
            },
            headers={"User-Agent": self._config.user_agent},
        )
        data = response.json()
        self.access_token = data["access_token"]
        self.refresh_token = data["refresh_token"]
        with set_pixiv_client(self):
            self.account = User.parse_obj(data["user"])
        return self.account

    async def login(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
    ) -> User:
        if token is not None:
            return await self.login_with_token(token)
        else:
            return await self.login_with_pwd(username, password)

    async def download(
        self,
        url: StrOrURL,
        *,
        output: Optional[Union[str, Path, BytesIO]] = None,
        chunk_size: Optional[int] = None,
    ) -> bytes:
        data = b""
        async with self.stream("GET", url) as response:
            if not isinstance(output, BytesIO) and output:
                output = Path(output).resolve()

                async with async_open(output) as file:
                    async for chunk in response.aiter_bytes(chunk_size):
                        data += chunk
                        await file.write(chunk)
            else:
                async for chunk in response.aiter_bytes(chunk_size):
                    data += chunk
        return data



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    AsyncIterator,
    Generic,
    Iterator,
    List,
    Optional,
    TypeVar,
)

from pydantic import (
    AnyHttpUrl,
    Field,
)
from typing_extensions import Self

from async_pixiv.model._base import (
    PixivModel,
    null_dict_validator,
)
from async_pixiv.model.illust import Comment, Illust, UgoiraMetadata
from async_pixiv.model.novel import (
    Novel,
    NovelMaker,
    NovelSeries,
)
from async_pixiv.model.user import (
    User,
    UserProfile,
    UserProfilePublicity,
    UserWorkSpace,
)
from async_pixiv.utils.context import set_pixiv_client

__all__ = [
    "UserPreview",
    "UserSearchResult",
    "UserIllustsResult",
    "UserBookmarksIllustsResult",
    "UserNovelsResult",
    "UserDetailResult",
    "UserRelatedResult",
    "IllustSearchResult",
    "IllustDetailResult",
    "IllustCommentResult",
    "IllustRelatedResult",
    "IllustNewResult",
    "UgoiraMetadataResult",
    "RecommendedResult",
    "NovelSearchResult",
    "NovelContentResult",
    "NovelSeriesResult",
    "NovelDetailResult",
]

T = TypeVar("T")


class PageResult(ABC, PixivModel, Generic[T]):
    next_url: Optional[AnyHttpUrl]
    search_span_limit: Optional[int]

    @abstractmethod
    def __iter__(self) -> Iterator[T]:
        pass

    async def next(self) -> Optional[Self]:
        if getattr(self, "next_url", None):
            data = (await self._pixiv_client.get(self.next_url)).json()
            with set_pixiv_client(self._pixiv_client):
                return self.__class__.parse_obj(data)
        else:
            return None

    async def iter_all_pages(self) -> AsyncIterator[T]:
        for result in self:
            yield result
        next_results = await self.next()
        while next_results is not None:
            for result in next_results:
                yield result
            next_results = await next_results.next()


class UserPreview(PixivModel):
    user: User
    illusts: List[Illust]
    is_muted: bool


class UserSearchResult(PageResult[UserPreview]):
    users: List[UserPreview] = Field([], alias="user_previews")

    def __iter__(self) -> Iterator[UserPreview]:
        return iter(self.users)


class UserIllustsResult(PageResult[Illust]):
    illusts: List[Illust]

    def __iter__(self) -> Iterator[Illust]:
        return iter(self.illusts)


class UserBookmarksIllustsResult(UserIllustsResult):
    pass


class UserNovelsResult(PageResult[Novel]):
    novels: List[Novel] = []

    def __iter__(self) -> Iterator[Novel]:
        return iter(self.novels)


class UserDetailResult(PixivModel):
    user: User
    profile: UserProfile
    profile_publicity: UserProfilePublicity
    workspace: UserWorkSpace


class UserRelatedResult(PixivModel):
    users: List[UserPreview] = Field(alias="user_previews")

    def __iter__(self) -> Iterator[UserPreview]:
        return iter(self.users)


class IllustSearchResult(UserIllustsResult):
    pass


class IllustDetailResult(PixivModel):
    illust: Illust


class IllustCommentResult(PageResult[Comment]):
    total: int = Field(0, alias="total_comments")
    comments: List[Comment] = []
    next_url: Optional[AnyHttpUrl]
    access_comment: Optional[bool] = Field(alias="comment_access_control")

    def __iter__(self) -> Iterator[Comment]:
        return iter(self.comments)


class IllustRelatedResult(IllustSearchResult):
    pass


class IllustNewResult(IllustDetailResult):
    pass


class UgoiraMetadataResult(PixivModel):
    metadata: UgoiraMetadata = Field(alias="ugoira_metadata")


class RecommendedResult(IllustSearchResult):
    ranking_illusts: List[Illust]
    contest_exists: bool


class NovelSearchResult(UserNovelsResult):
    pass


class NovelDetailResult(PixivModel):
    novel: Novel


class NovelContentResult(PixivModel):
    marker: Optional[NovelMaker] = Field(alias="novel_marker")
    content: str = Field(alias="novel_text")
    previous: Optional[Novel] = Field(alias="series_prev")
    next: Optional[Novel] = Field(alias="series_next")

    _check = null_dict_validator("marker", "previous", "next")

    @property
    def text(self) -> str:
        return self.content


class NovelSeriesResult(NovelSearchResult):
    series: NovelSeries = Field(alias="novel_series_detail")
    first_novel: Novel = Field(alias="novel_series_first_novel")
    latest_novel: Novel = Field(alias="novel_series_latest_novel")



from abc import ABC
from typing import (
    Dict,
    Optional,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

from typing_extensions import Literal
from yarl import URL

from async_pixiv.utils.enums import (
    Enum,
    SearchDuration,
    SearchFilter,
    SearchShort,
    SearchTarget,
)

if TYPE_CHECKING:
    from async_pixiv.client._client import PixivClient

API_HOST = URL("https://app-api.pixiv.net")
V1_API = API_HOST / "v1"
V2_API = API_HOST / "v2"


# noinspection PyShadowingBuiltins
class _Section(ABC):
    _client: "PixivClient"
    _type: str

    @property
    def type(self) -> str:
        return self._type

    def __init_subclass__(cls, **kwargs) -> None:
        cls._type = cls.__name__.lower()

    def __init__(self, client: "PixivClient") -> None:
        self._client = client

    async def search(
        self,
        word: str,
        *,
        sort: Union[
            Literal["date_desc", "date_asc", "popular_desc", "popular_asc"], SearchShort
        ] = SearchShort.date_desc,
        target: Union[
            SearchTarget,
            Literal[
                "partial_match_for_tags", "keyword", "exact_match_for_tags", "text"
            ],
        ] = None,
        duration: Optional[
            Union[
                Literal[
                    "within_last_day",
                    "within_last_week",
                    "within_last_month",
                    "within_last_year",
                ],
                SearchDuration,
            ]
        ] = None,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
        offset: Optional[int] = None,
        **kwargs,
    ) -> Dict:
        response = await self._client.get(
            V1_API / f"search/{self._type}",
            params={
                "word": word,
                "sort": sort,
                "duration": duration,
                "filter": filter,
                "offset": offset,
                "search_target": target,
                **kwargs,
            },
        )
        return response.json()

    async def detail(
        self,
        id: int,
        *,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
    ) -> Dict:
        request = await self._client.get(
            V1_API / f"{self._type}/detail",
            params={f"{self._type}_id": id, "filter": filter},
        )
        return request.json()


SectionType = TypeVar("SectionType", bound=_Section)


class IllustType(Enum):
    illust = "illust"
    manga = "manga"
    ugoira = "ugoira"
    novel = "novel"



from datetime import date
from typing import (
    Optional,
    Union,
)

from typing_extensions import Literal

from async_pixiv.client._section._base import (
    SearchDuration,
    SearchFilter,
    SearchShort,
    SearchTarget,
    V1_API,
    V2_API,
    _Section,
)
from async_pixiv.model.result import (
    NovelContentResult,
    NovelDetailResult,
    NovelSearchResult,
    NovelSeriesResult,
)

__all__ = ("NOVEL",)

from async_pixiv.utils.context import set_pixiv_client


# noinspection PyShadowingBuiltins
class NOVEL(_Section):
    async def search(
        self,
        word: str,
        *,
        sort: Union[
            Literal["date_desc", "date_asc", "popular_desc", "popular_asc"], SearchShort
        ] = SearchShort.date_desc,
        target: Union[
            SearchTarget,
            Literal[
                "partial_match_for_tags", "keyword", "exact_match_for_tags", "text"
            ],
        ] = SearchTarget.partial,
        duration: Optional[
            Union[
                Literal[
                    "within_last_day",
                    "within_last_week",
                    "within_last_month",
                    "within_last_year",
                ],
                SearchDuration,
            ]
        ] = None,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
        offset: Optional[int] = None,
        min_bookmarks: Optional[int] = None,
        max_bookmarks: Optional[int] = None,
        start_date: Optional[date] = None,
        end_date: Optional[date] = None,
        merge_plain_keyword_results: bool = True,
        include_translated_tag_results: bool = True,
        **kwargs,
    ) -> NovelSearchResult:
        if start_date is not None:
            start_date = start_date.strftime("%Y-%m-%d")
        if end_date is not None:
            end_date = end_date.strftime("%Y-%m-%d")
        data = await super(NOVEL, self).search(
            word,
            sort=sort,
            target=target,
            duration=duration,
            filter=filter,
            offset=offset,
            bookmark_num_min=min_bookmarks,
            bookmark_num_max=max_bookmarks,
            start_date=start_date,
            end_date=end_date,
            merge_plain_keyword_results=merge_plain_keyword_results,
            include_translated_tag_results=include_translated_tag_results,
            **kwargs,
        )
        with set_pixiv_client(self._client):
            return NovelSearchResult.parse_obj(data)

    async def detail(
        self,
        id: int,
        *,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
    ) -> NovelDetailResult:
        data = (
            await self._client.get(V2_API / "novel/detail", params={"novel_id": id})
        ).json()
        with set_pixiv_client(self._client):
            return NovelDetailResult.parse_obj(data)

    async def content(self, id: int) -> NovelContentResult:
        data = (
            await self._client.get(V1_API / "novel/text", params={"novel_id": id})
        ).json()
        with set_pixiv_client(self._client):
            return NovelContentResult.parse_obj(data)

    async def series(
        self,
        id: int,
        *,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
    ) -> NovelSeriesResult:
        data = (
            await self._client.get(
                V2_API / "novel/series", params={"series_id": id, "filter": filter}
            )
        ).json()
        with set_pixiv_client(self._client):
            return NovelSeriesResult.parse_obj(data)



import asyncio
import collections
from typing import List, Optional, Tuple

from async_pixiv.utils.ffmpeg.typing import Option

try:
    import regex as re
except ImportError:
    import re

__all__ = (
    "Progress",
    "progress_pattern",
    "build_option",
    "build_options",
    "read_lines",
    "parse_progress",
)

Progress = collections.namedtuple(
    "Progress", ["frame", "fps", "size", "time", "bitrate", "speed"]
)

progress_pattern = re.compile(r"(frame|fps|size|time|bitrate|speed)\s*\=\s*(\S+)")


def build_option(key: str, value: Option) -> List[str]:
    if key.startswith("-"):
        key = key[1:]

    option = [f"-{key}"]
    if value is not None:
        option.append(str(value))

    return option


def build_options(options: List[Tuple[str, Option]]) -> List[str]:
    arguments = []

    for key, values in options:
        if not isinstance(values, (list, set, tuple)):
            values = [values]

        for value in values:
            arguments.extend(build_option(key, value))

    return arguments


async def read_lines(stream: asyncio.StreamReader):
    pattern = re.compile(rb"[\r\n]+")

    data = bytearray()
    while not stream.at_eof():
        lines = pattern.split(data)
        data[:] = lines.pop(-1)

        for line in lines:
            yield line

        data.extend(await stream.read(1024))


# Reference: https://github.com/FFmpeg/FFmpeg/blob/master/fftools/ffmpeg.c#L1646
def parse_progress(line: str) -> Optional[Progress]:
    default = {
        "frame": "0",
        "fps": "0.0",
        "size": "0kB",
        "time": "00:00:00.00",
        "bitrate": "0.0kbits/s",
        "speed": "0.0x",
    }

    items = {
        key: value for key, value in progress_pattern.findall(line) if value != "N/A"
    }

    if not items:
        return None

    progress = {
        **default,
        **items,
    }

    return Progress(
        frame=int(progress["frame"]),
        fps=float(progress["fps"]),
        size=int(progress["size"].replace("kB", "")) * 1024,
        time=progress["time"],
        bitrate=float(progress["bitrate"].replace("kbits/s", "")),
        speed=float(progress["speed"].replace("x", "")),
    )



import asyncio
from contextlib import asynccontextmanager
from logging import getLogger
from typing import Any, AsyncIterator, Optional, Union

from aiolimiter import AsyncLimiter

# noinspection PyProtectedMember
from httpx import AsyncClient, URL

# noinspection PyProtectedMember
from httpx._client import USE_CLIENT_DEFAULT, UseClientDefault

# noinspection PyProtectedMember
from httpx._config import Proxy

# noinspection PyProtectedMember
from httpx._types import (
    AuthTypes,
    CookieTypes,
    HeaderTypes,
    ProxiesTypes,
    QueryParamTypes,
    RequestContent,
    RequestData,
    RequestExtensions,
    RequestFiles,
    TimeoutTypes,
    URLTypes,
)

# noinspection PyProtectedMember
from httpx._utils import get_environment_proxies
from yarl import URL

from async_pixiv.utils.context import async_do_nothing
from async_pixiv.utils.overwrite import AsyncHTTPTransport, Response

try:
    import regex as re
except ImportError:
    import re

try:
    from uvloop import EventLoopPolicy

    asyncio.set_event_loop_policy(EventLoopPolicy())
except ImportError:
    EventLoopPolicy = None

__all__ = ["Net"]

logger = getLogger(__name__)


class Net:
    _client: AsyncClient

    @property
    def client(self) -> AsyncClient:
        return self._client

    # noinspection PyProtectedMember
    def __init__(
        self,
        *,
        rate_limiter: Optional[AsyncLimiter] = AsyncLimiter(100),
        proxies: Optional[ProxiesTypes] = None,
        retry: Optional[int] = 5,
        retry_sleep: float = 1,
        **kwargs,
    ):
        self._rate_limiter = rate_limiter
        if proxies is None:
            proxy_map = {
                key: None if url is None else Proxy(url=url)
                for key, url in get_environment_proxies().items()
            }
        elif isinstance(proxies, dict):
            new_proxies = {}
            for key, value in proxies.items():
                proxy = Proxy(url=value) if isinstance(value, (str, URL)) else value
                new_proxies[str(key)] = proxy
            proxy_map = new_proxies
        else:
            proxy = Proxy(url=proxies) if isinstance(proxies, (str, URL)) else proxies
            proxy_map = {"all://": proxy}

        for i in filter(lambda x: x in kwargs, ["mounts", "transport", "proxies"]):
            del kwargs[i]
        self._retry = retry
        self._retry_sleep = max(retry_sleep, 0)
        self._client = AsyncClient(
            mounts={k: AsyncHTTPTransport(proxy=v) for k, v in proxy_map.items()},
            transport=AsyncHTTPTransport(),
            **kwargs,
        )

    async def request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
        cookies: Optional[CookieTypes] = None,
        auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: Union[UseClientDefault, bool] = USE_CLIENT_DEFAULT,
        timeout: Union[UseClientDefault, TimeoutTypes] = USE_CLIENT_DEFAULT,
        extensions: Optional[RequestExtensions] = None,
    ) -> Response:
        request = self._client.build_request(
            method=method,
            url=url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            timeout=timeout,
            extensions=extensions,
        )
        error = None
        times = max(1, self._retry)
        for n in range(times):
            try:
                async with (self._rate_limiter or async_do_nothing()):
                    return await self._client.send(
                        request, auth=auth, follow_redirects=follow_redirects
                    )
            except Exception as e:
                error = e
                logger.warning(
                    f"Request Error: {e}. Will retry in {self._retry_sleep}s. ({n + 1}th retry)"
                )
                await asyncio.sleep(self._retry_sleep)
        if error is not None:
            raise error

    async def get(
        self,
        url: URLTypes,
        *,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
        cookies: Optional[CookieTypes] = None,
        auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: Union[UseClientDefault, bool] = USE_CLIENT_DEFAULT,
        timeout: Union[UseClientDefault, TimeoutTypes] = USE_CLIENT_DEFAULT,
        extensions: Optional[RequestExtensions] = None,
    ) -> Response:
        return await self.request(
            "GET",
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )

    async def post(
        self,
        url: URLTypes,
        *,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
        cookies: Optional[CookieTypes] = None,
        auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: Union[UseClientDefault, bool] = USE_CLIENT_DEFAULT,
        timeout: Union[UseClientDefault, TimeoutTypes] = USE_CLIENT_DEFAULT,
        extensions: Optional[RequestExtensions] = None,
    ) -> Response:
        return await self.request(
            "POST",
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )

    @asynccontextmanager
    async def stream(
        self,
        method: str,
        url: URLTypes,
        *,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
        cookies: Optional[CookieTypes] = None,
        auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: Union[UseClientDefault, bool] = USE_CLIENT_DEFAULT,
        timeout: Union[UseClientDefault, TimeoutTypes] = USE_CLIENT_DEFAULT,
        extensions: Optional[RequestExtensions] = None,
    ) -> AsyncIterator[Response]:
        request = self._client.build_request(
            method=method,
            url=url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            timeout=timeout,
            extensions=extensions,
        )
        response = await self._client.send(
            request=request,
            auth=auth,
            follow_redirects=follow_redirects,
            stream=True,
        )
        try:
            yield response
        finally:
            await response.aclose()

    async def close(self) -> None:
        return await self._client.aclose()



from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from async_pixiv.utils.overwrite import Response


class PixivError(Exception):
    def __init_subclass__(cls, **kwargs) -> None:
        cls.__module__ = "async_pixiv.error"


class LoginError(PixivError):
    pass


class OauthError(PixivError):
    pass


class ArtWorkTypeError(PixivError, TypeError):
    pass


class StatusError(PixivError):
    reason: Optional[str] = None

    def __init__(self, response: "Response") -> None:
        self.response = response

    def __str__(self) -> str:
        return self.reason or self.response


class ApiError(PixivError):
    user_message: Optional[str] = None
    message: Optional[str] = None
    reason: Optional[str] = None
    user_message_details: Optional[str] = None

    def __init__(self, data: dict) -> None:
        self.user_message = data.get("user_message", self.user_message)
        self.message = data.get("message", self.message)
        self.reason = data.get("reason", self.reason)
        self.user_message_details = data.get(
            "user_message_details", self.user_message_details
        )

    def __str__(self) -> str:
        return (
            self.user_message
            or self.message
            or self.reason
            or self.user_message_details
        )


class RateLimit(ApiError):
    reason = "速率限制"


class NotExist(StatusError):
    reason = "作品不存在或无浏览权限"



import asyncio
import signal
import sys
from codecs import StreamReader
from collections import namedtuple
from typing import Any, Coroutine, Dict, List, Optional, TYPE_CHECKING

from pyee import EventEmitter

from async_pixiv.utils.ffmpeg.typing import Option
from async_pixiv.utils.ffmpeg.utils import build_options, parse_progress, read_lines

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from asyncio.subprocess import Process

_windows = sys.platform == "win32"


def _create_subprocess(*args, **kwargs) -> Coroutine[Any, Any, "Process"]:
    if _windows:
        # https://docs.python.org/3/library/asyncio-subprocess.html#asyncio.asyncio.subprocess.Process.send_signal
        from subprocess import CREATE_NEW_PROCESS_GROUP

        kwargs["creationflags"] = CREATE_NEW_PROCESS_GROUP

    return asyncio.create_subprocess_exec(*args, **kwargs)


class FFmpegError(Exception):
    pass


class FFmpeg(EventEmitter):
    _File = namedtuple("_File", ["url", "options"])

    def __init__(self, executable: str = "ffmpeg", skip_default_drain: bool = False):
        super().__init__()

        self._executable = executable
        self._global_options = []
        self._input_files = []
        self._output_files = []

        self._executed = False
        self._terminated = False
        self._process = None
        self._skip_default_drain = skip_default_drain

        self.on("stderr", self._on_stderr)

    def option(self, key: str, value: Optional[Option] = None) -> "FFmpeg":
        self._global_options.append((key, value))
        return self

    def input(
        self, url: str, options: Optional[Dict[str, Option]] = None, **kwargs
    ) -> "FFmpeg":
        if options is None:
            options = {}

        self._input_files.append(FFmpeg._File(url=url, options={**options, **kwargs}))
        return self

    def output(
        self, url: str, options: Optional[Dict[str, Option]] = None, **kwargs
    ) -> "FFmpeg":
        if options is None:
            options = {}

        self._output_files.append(FFmpeg._File(url=url, options={**options, **kwargs}))
        return self

    async def execute(self, stream: Optional[asyncio.StreamReader] = None):
        if self._executed:
            raise FFmpegError("FFmpeg is already executed")

        arguments = self._build()
        self.emit("start", arguments)

        self._process = await _create_subprocess(
            *arguments,
            stdin=asyncio.subprocess.PIPE if stream is not None else None,
            stderr=asyncio.subprocess.PIPE,
        )

        self._executed = True
        await asyncio.wait(
            [
                asyncio.create_task(self._write_stdin(stream)),
                asyncio.create_task(self._read_stderr()),
                asyncio.create_task(self._process.wait()),
            ]
        )

        if self._process.returncode == 0:
            self.emit("completed")
        elif self._terminated:
            self.emit("terminated")
        else:
            self.emit("error", self._process.returncode)

    def terminate(self):
        if not self._executed:
            raise FFmpegError("FFmpeg is not executed")

        sigterm = signal.SIGTERM
        if _windows:  # On Windows, SIGTERM -> TerminateProcess()
            # https://github.com/FFmpeg/FFmpeg/blob/master/fftools/ffmpeg.c#L356
            sigterm = signal.CTRL_BREAK_EVENT

        self._terminated = True
        self._process.send_signal(sigterm)

    async def drain(self):
        # Memory leak occurs when too much data is feed to StreamReader (stream)
        # So expe StreamWriter's (stdin) drain to be called to create back pressure
        return await self._process.stdin.drain()

    async def _write_stdin(self, stream: Optional[StreamReader]):
        if stream is None:
            return

        while not stream.at_eof():
            # noinspection PyUnresolvedReferences
            self._process.stdin.write(await stream.read(1024))
            if not self._skip_default_drain:
                await self._process.stdin.drain()
        self._process.stdin.write_eof()

    async def _read_stderr(self):
        async for line in read_lines(self._process.stderr):
            self.emit("stderr", line.decode("utf-8"))

    def _on_stderr(self, line: str):  # registered in __init__()
        progress = parse_progress(line)
        if progress:
            self.emit("progress", progress)

    def _build(self) -> List[str]:
        arguments = [self._executable]
        arguments.extend(build_options(self._global_options))

        for file in self._input_files:
            arguments.extend(build_options(file.options))
            arguments.extend(["-i", file.url])

        for file in self._output_files:
            arguments.extend(build_options(file.options))
            arguments.append(file.url)

        return arguments



from asyncio import Event
from datetime import datetime
from enum import Enum
from io import BytesIO
from pathlib import Path
from typing import Any, List, Optional, TYPE_CHECKING, Union, overload
from zipfile import ZipFile

from aiofiles import open as async_open

# noinspection PyUnresolvedReferences
from aiofiles.tempfile import TemporaryDirectory
from pydantic import (
    AnyHttpUrl,
    Field,
    PrivateAttr,
)
from requests import HTTPError, Session
from typing_extensions import Literal
from yarl import URL

from async_pixiv.error import ArtWorkTypeError
from async_pixiv.model._base import (
    PixivModel,
    null_dict_validator,
)
from async_pixiv.model.other import (
    AIType,
    ImageUrl,
    Series,
    Tag,
)
from async_pixiv.model.user import User
from async_pixiv.utils.ffmpeg import FFmpeg

try:
    import regex as re
except ImportError:
    import re

if TYPE_CHECKING:
    from async_pixiv.client import PixivClient
    from async_pixiv.model.result import (
        IllustDetailResult,
        IllustCommentResult,
        IllustRelatedResult,
    )

__all__ = [
    "Illust",
    "IllustType",
    "Comment",
    "UgoiraMetadata",
]

UGOIRA_RESULT_TYPE = Literal["zip", "jpg", "iter", "gif", "mp4"]

session = Session()


class IllustType(Enum):
    illust = "illust"
    ugoira = "ugoira"
    manga = "manga"
    novel = "novel"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, IllustType):
            return self.value == other.value
        else:
            try:
                return self.value == str(other)
            except (TypeError, ValueError):
                return False


# noinspection PyProtectedMember,PyShadowingBuiltins
class Illust(PixivModel):
    class MetaPage(PixivModel):
        image_urls: "ImageUrl"

    class MetaSinglePage(PixivModel):
        original: Optional[AnyHttpUrl] = Field(alias="original_image_url")

        @property
        def link(self):
            return self.original

    id: int
    title: str
    type: IllustType
    image_urls: "ImageUrl"
    caption: str
    restrict: int
    user: "User"
    tags: List[Tag]
    tools: List[str]
    create_date: datetime
    page_count: int
    width: int
    height: int
    sanity_level: int
    x_restrict: int
    series: Optional[Series]
    meta_single_page: MetaSinglePage
    meta_pages: List[MetaPage]
    total_view: int
    total_bookmarks: int
    is_bookmarked: bool
    visible: bool
    is_muted: bool
    total_comments: Optional[int]
    comment_access_control: Optional[int]
    ai_type: AIType = Field(alias="illust_ai_type")

    _ugoira_metadata: Optional["UgoiraMetadata"] = PrivateAttr(None)
    _is_r18: Optional[bool] = PrivateAttr(None)
    _is_r18g: Optional[bool] = PrivateAttr(None)

    @property
    def link(self) -> URL:
        return URL(f"https://www.pixiv.net/artworks/{self.id}/")

    @property
    def is_nsfw(self) -> bool:
        return self.sanity_level > 5

    async def is_r18(self) -> bool:
        if self._is_r18 is None:
            try:
                client = self._pixiv_client
                response = await client.get(
                    f"https://www.pixiv.net/ajax/illust/{self.id}",
                    follow_redirects=True,
                )
                response.raise_for_status()
                self._is_r18 = "R-18" in list(
                    map(lambda x: x["tag"], response.json()["body"]["tags"]["tags"])
                )
            except HTTPError:
                self._is_r18 = any(
                    map(
                        lambda x: ("R-18" in x.name.upper() or "R18" in x.name.upper()),
                        self.tags,
                    )
                )

        return self._is_r18

    async def is_r18g(self) -> bool:
        if self._is_r18g is None:
            try:
                client = self._pixiv_client
                response = await client.get(self.link, follow_redirects=True)
                response.raise_for_status()
                html = response.text
                title = re.findall(
                    r"<meta property=\"twitter:title\" content=\"(.*?)\">", html
                )[0]
                self._is_r18g = title.startswith("[R-18G]")
            except HTTPError:
                self._is_r18g = any(
                    map(
                        lambda x: (
                            "R-18G" in x.name.upper() or "R18G" in x.name.upper()
                        ),
                        self.tags,
                    )
                )
        return self._is_r18g

    @property
    def all_image_urls(self) -> List[URL]:
        if not self.meta_pages:
            # noinspection PyTypeChecker
            return [self.meta_single_page.original]
        result = []
        for page in self.meta_pages:
            result.append(URL(page.image_urls.link))
        return result

    async def detail(self, *, for_ios: bool = True) -> "IllustDetailResult":
        from async_pixiv.client._section._base import SearchFilter

        return await self._pixiv_client.ILLUST.detail(
            self.id, filter=SearchFilter.ios if for_ios else SearchFilter.android
        )

    async def comments(self, *, offset: Optional[int] = None) -> "IllustCommentResult":
        return await self._pixiv_client.ILLUST.comments(self.id, offset=offset)

    async def related(
        self,
        *,
        for_ios: bool = True,
        offset: Optional[int] = None,
        seed_id: Optional[int] = None,
    ) -> "IllustRelatedResult":
        from async_pixiv.client._section._base import SearchFilter

        return await self._pixiv_client.ILLUST.related(
            self.id,
            offset=offset,
            seed_ids=seed_id,
            filter=SearchFilter.ios if for_ios else SearchFilter.android,
        )

    async def download(
        self,
        *,
        full: bool = False,
        output: Optional[Union[str, Path]] = None,
        client: Optional["PixivClient"] = None,
    ) -> List[bytes]:
        if client is None:
            from async_pixiv.client import PixivClient

            client = PixivClient.get_client()
        if not full or not self.meta_pages:
            return [
                await client.download(
                    self.meta_single_page.link or self.image_urls.link
                )
            ]
        else:
            result: List[bytes] = []
            for meta_page in self.meta_pages:
                result.append(
                    await client.download(meta_page.image_urls.link, output=output)
                )
            return result

    async def ugoira_metadata(self) -> Optional["UgoiraMetadata"]:
        if self.type != IllustType.ugoira:
            return None
        if self._ugoira_metadata is None:
            self._ugoira_metadata = (
                await self._pixiv_client.ILLUST.ugoira_metadata(self.id)
            ).metadata
        return self._ugoira_metadata

    @overload
    async def download_ugoira(self, *, type: Literal["zip"]) -> Optional[bytes]:
        pass

    @overload
    async def download_ugoira(self, *, type: Literal["all"]) -> Optional[List[bytes]]:
        pass

    @overload
    async def download_ugoira(self, *, type: Literal["gif"]) -> Optional[bytes]:
        pass

    @overload
    async def download_ugoira(self, *, type: Literal["mp4"]) -> Optional[bytes]:
        pass

    async def download_ugoira(
        self, *, type: UGOIRA_RESULT_TYPE = "zip"
    ) -> Union[bytes, List[bytes], None]:
        if self.type != IllustType.ugoira:
            raise ArtWorkTypeError(
                "If you want to download a normal image, "
                'please use this method: "download"'
            )
        metadata = await self.ugoira_metadata()
        data = await self._pixiv_client.download(metadata.zip_url.link)
        if data is None:
            return None
        if type == "zip":
            return data

        zip_file = ZipFile(BytesIO(data))

        if type == "all":
            frames = []
            for frame in metadata.frames:
                with zip_file.open(frame.file) as f:
                    frames.append(f.read())
            return frames

        async with TemporaryDirectory() as directory:
            directory = Path(directory).resolve()
            connect_config_file_path = directory / "list.txt"
            async with async_open(connect_config_file_path, mode="w") as list_file:
                for frame in metadata.frames:
                    with zip_file.open(frame.file) as frame_zip_file:
                        frame_file_path = directory / frame.file
                        async with async_open(frame_file_path, mode="wb") as frame_file:
                            await frame_file.write(frame_zip_file.read())
                    await list_file.write(
                        f"file {frame_file_path.resolve()}\n".replace("\\", "/")
                    )
                    await list_file.write(f"duration {frame.delay / 1000}\n")
            del zip_file
            event = Event()
            if type == "mp4":
                output_path = directory / "out.mp4"
                ffmpeg = (
                    FFmpeg()
                    .option("y")
                    .option("f", "lavfi")
                    .option("i", "anullsrc")
                    .option("f", "concat")
                    .option("safe", 0)
                    .option(
                        "filter_complex",
                        "crop='iw-mod(iw,2)':'ih-mod(ih,2)'[main];"
                        "[main]split[v1][v2];"
                        "[v1]palettegen[pal];"
                        "[v2][pal]paletteuse=dither=sierra2_4a",
                    )
                    .option("i", str(connect_config_file_path.resolve()))
                    .option("pix_fmt", "yuv420p10le")
                    .option("c:v", "libx265")
                    .option("c:a", "aac")
                    .option("crf", 0)
                    .option("x265-params", "profile=main10")
                    .option("shortest")
                    .output(str(output_path))
                )
            else:
                output_path = directory / "out.gif"
                ffmpeg = (
                    FFmpeg()
                    .option("y")
                    .option("f", "concat")
                    .option("safe", 0)
                    .option("i", str(connect_config_file_path.resolve()))
                    .option(
                        "filter_complex",
                        "split[v1][v2];"
                        "[v1]palettegen[pal];"
                        "[v2][pal]paletteuse=dither=sierra2_4a",
                    )
                    .option("crf", 0)
                    .output(str(output_path))
                )
            ffmpeg.on("completed", lambda: event.set())
            await ffmpeg.execute()
            await event.wait()

            async with async_open(output_path, mode="rb") as file:
                return await file.read()


class Comment(PixivModel):
    id: int
    comment: str
    date: datetime
    user: User
    parent: Optional["Comment"] = Field(alias="parent_comment")

    _check = null_dict_validator("parent")


class UgoiraMetadata(PixivModel):
    class Frame(PixivModel):
        file: str
        delay: int

    zip_url: ImageUrl = Field(alias="zip_urls")
    frames: List[Frame]



"""重写 telegram.request.HTTPXRequest 使其使用 ujson 库进行 json 序列化"""
from typing import Any, AsyncIterable, Iterable

import httpcore
from httpx import (
    AsyncByteStream,
    AsyncHTTPTransport as DefaultAsyncHTTPTransport,
    HTTPTransport as DefaultHTTPTransport,
    Request,
    Response as DefaultResponse,
    SyncByteStream,
)

# noinspection PyProtectedMember
from httpx._transports.default import (
    AsyncResponseStream,
    ResponseStream,
    map_httpcore_exceptions,
)

# noinspection PyProtectedMember
from httpx._utils import guess_json_utf

from async_pixiv.error import ApiError, NotExist, RateLimit

try:
    import ujson as jsonlib
except ImportError:
    import json as jsonlib

__all__ = ("AsyncHTTPTransport", "HTTPTransport", "Response")

STATUS_ERROR_MAP = {
    404: NotExist,
}
RESULT_ERROR_MAP = {
    "Rate Limit": RateLimit,
}


class Response(DefaultResponse):
    def raise_for_status(self) -> None:
        if self.status_code != 200:
            if (error := STATUS_ERROR_MAP.get(self.status_code)) is not None:
                raise error(self)
            else:
                super().raise_for_status()

    # noinspection PyMethodMayBeStatic
    def raise_for_result(self, result: dict) -> None:
        if (error := result.get("error")) is not None and error:
            raise RESULT_ERROR_MAP.get(error["reason"], ApiError)(error)

    def json(
        self,
        raise_for_result: bool = True,
        raise_for_status: bool = True,
        **kwargs: Any,
    ) -> Any:
        if raise_for_status:
            self.raise_for_status()

        result = None
        if self.charset_encoding is None and self.content and len(self.content) > 3:
            encoding = guess_json_utf(self.content)
            if encoding is not None:
                result = jsonlib.loads(self.content.decode(encoding), **kwargs)

        result = jsonlib.loads(self.text, **kwargs) if result is None else result

        if isinstance(result, dict) and raise_for_result:
            self.raise_for_result(result)

        return result


# noinspection PyProtectedMember
class AsyncHTTPTransport(DefaultAsyncHTTPTransport):
    async def handle_async_request(self, request: Request) -> Response:
        assert isinstance(request.stream, AsyncByteStream)

        req = httpcore.Request(
            method=request.method,
            url=httpcore.URL(
                scheme=request.url.raw_scheme,
                host=request.url.raw_host,
                port=request.url.port,
                target=request.url.raw_path,
            ),
            headers=request.headers.raw,
            content=request.stream,
            extensions=request.extensions,
        )
        with map_httpcore_exceptions():
            resp = await self._pool.handle_async_request(req)

        assert isinstance(resp.stream, AsyncIterable)

        return Response(
            status_code=resp.status,
            headers=resp.headers,
            stream=AsyncResponseStream(resp.stream),
            extensions=resp.extensions,
        )


class HTTPTransport(DefaultHTTPTransport):
    # noinspection PyProtectedMember
    def handle_request(self, request: Request) -> Response:
        assert isinstance(request.stream, SyncByteStream)

        req = httpcore.Request(
            method=request.method,
            url=httpcore.URL(
                scheme=request.url.raw_scheme,
                host=request.url.raw_host,
                port=request.url.port,
                target=request.url.raw_path,
            ),
            headers=request.headers.raw,
            content=request.stream,
            extensions=request.extensions,
        )
        with map_httpcore_exceptions():
            resp = self._pool.handle_request(req)

        assert isinstance(resp.stream, Iterable)

        return Response(
            status_code=resp.status,
            headers=resp.headers,
            stream=ResponseStream(resp.stream),
            extensions=resp.extensions,
        )



from datetime import datetime
from typing import (
    List,
    Optional,
    TYPE_CHECKING,
)

from pydantic import (
    Extra,
    Field,
    PrivateAttr,
)
from requests import HTTPError
from yarl import URL

from async_pixiv.model._base import (
    PixivModel,
    PixivModelConfig,
    null_dict_validator,
)
from async_pixiv.model.other import (
    AIType,
    ImageUrl,
    Series,
    Tag,
)
from async_pixiv.model.user import User

try:
    import regex as re
except ImportError:
    import re

if TYPE_CHECKING:
    from async_pixiv.client import PixivClient
    from async_pixiv.model.result import (
        NovelContentResult,
        NovelDetailResult,
        NovelSeriesResult,
    )

__all__ = ["Novel", "NovelMaker", "NovelSeries"]


# noinspection PyProtectedMember
class Novel(PixivModel):
    id: int
    title: str
    caption: str
    restrict: int
    x_restrict: int
    is_original: bool
    image_url: ImageUrl = Field(alias="image_urls")
    create_date: datetime
    tags: List[Tag]
    page_count: int
    text_length: int
    user: User
    series: Optional[Series]
    is_bookmarked: bool
    total_bookmarks: int
    total_view: int
    visible: bool
    total_comments: int
    is_muted: bool
    is_mypixiv_only: bool
    is_x_restricted: bool
    comment_access_control: Optional[int]
    ai_type: AIType = Field(alias="novel_ai_type")

    _check = null_dict_validator("series")
    _is_r18: Optional[bool] = PrivateAttr(None)
    _lang: Optional[str] = PrivateAttr(None)

    @property
    def link(self) -> URL:
        return URL(f"https://www.pixiv.net/novel/show.php?id={self.id}")

    async def is_r18(self) -> bool:
        if self._is_r18 is None:
            try:
                client = self._pixiv_client
                response = await client.get(self.link, follow_redirects=True)
                response.raise_for_status()
                html = response.text
                title: str = re.findall(
                    r"<meta property=\"twitter:title\" content=\"(.*?)\">", html
                )[0]
                self._is_r18 = title.startswith("[R-18")
            except HTTPError:
                self._is_r18 = any(
                    map(
                        lambda x: ("R-18" in x.name.upper() or "R18" in x.name.upper()),
                        self.tags,
                    )
                )

        return self._is_r18

    async def get_lang(self) -> str:
        if self._lang is None:
            try:
                client = self._pixiv_client
                response = await client.get(
                    self.link,
                    follow_redirects=True,
                    headers={
                        "user-agent": (
                            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                            "AppleWebKit/537.36 (KHTML, like Gecko) "
                            "Chrome/108.0.0.0 Safari/537.36"
                        ),
                    },
                )
                response.raise_for_status()
                html = response.text
                self._lang = re.findall(r"\"language\":\"(.*?)\",", html)[0]
            except HTTPError:
                pass
        return self._lang

    async def detail(self, *, for_ios: bool = True) -> "NovelDetailResult":
        from async_pixiv.client._section._base import SearchFilter

        return await self._pixiv_client.NOVEL.detail(
            self.id, filter=SearchFilter.ios if for_ios else SearchFilter.android
        )

    async def content(
        self, client: Optional["PixivClient"] = None
    ) -> "NovelContentResult":
        if client is None:
            from async_pixiv.client import PixivClient

            client = PixivClient.get_client()
        return await client.NOVEL.content(self.id)

    async def series_detail(
        self, client: Optional["PixivClient"] = None
    ) -> Optional["NovelSeriesResult"]:
        if self.series is None:
            return None
        if client is None:
            from async_pixiv.client import PixivClient

            client = PixivClient.get_client()
        return await client.NOVEL.series(self.series.id)

    def __eq__(self, other: "Novel") -> bool:
        if not isinstance(other, Novel):
            return False
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(f"pixiv_novel_{self.id}")


class NovelMaker(PixivModel):
    page: int

    class Config(PixivModelConfig):
        extra = Extra.allow


class NovelSeries(Series):
    caption: str
    is_original: bool
    is_concluded: bool
    character_count: int = Field(alias="total_character_count")
    user: User
    display: str = Field(alias="display_text")
    watchlist_added: bool



from os import PathLike
from typing import TYPE_CHECKING, Union

__all__ = ("StrPath",)

if TYPE_CHECKING:
    from pathlib import Path

    StrPath = Union[str, PathLike[str], Path]
else:
    StrPath = Union[str, PathLike[str], "Path"]



from typing import (
    Optional,
    Union,
)

from typing_extensions import Literal

from async_pixiv.client._section._base import (
    SearchDuration,
    SearchFilter,
    SearchShort,
    V1_API,
    _Section,
)
from async_pixiv.model.illust import IllustType
from async_pixiv.model.result import (
    UserBookmarksIllustsResult,
    UserDetailResult,
    UserIllustsResult,
    UserNovelsResult,
    UserRelatedResult,
    UserSearchResult,
)

__all__ = ("USER",)

from async_pixiv.utils.context import set_pixiv_client


# noinspection PyShadowingBuiltins
class USER(_Section):
    async def search(
        self,
        word: str,
        *,
        sort: Union[
            Literal["date_desc", "date_asc", "popular_desc", "popular_asc"], SearchShort
        ] = SearchShort.date_desc,
        duration: Optional[
            Union[
                Literal[
                    "within_last_day",
                    "within_last_week",
                    "within_last_month",
                    "within_last_year",
                ],
                SearchDuration,
            ]
        ] = None,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
        offset: Optional[int] = None,
        **kwargs,
    ) -> UserSearchResult:
        data = await super(USER, self).search(
            word=word,
            sort=sort,
            duration=duration,
            filter=filter,
            offset=offset,
            **kwargs,
        )
        with set_pixiv_client(self._client):
            return UserSearchResult.parse_obj(data)

    async def detail(
        self,
        id: int,
        *,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
    ) -> UserDetailResult:
        if id is None:
            id = self._client.account.id
        data = await super(USER, self).detail(id=id, filter=filter)
        with set_pixiv_client(self._client):
            return UserDetailResult.parse_obj(data)

    async def illusts(
        self,
        id: Optional[int] = None,
        *,
        type: Union[
            Literal["illust", "manga", "novel", "ugoira"], IllustType
        ] = IllustType.illust,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
        offset: Optional[int] = None,
    ) -> UserIllustsResult:
        if id is None:
            id = self._client.account.id
        data = (
            await self._client.get(
                V1_API / "user/illusts",
                params={
                    "user_id": id,
                    "type": type,
                    "filter": filter,
                    "offset": offset,
                },
            )
        ).json()
        with set_pixiv_client(self._client):
            return UserIllustsResult.parse_obj(data)

    async def bookmarks(
        self,
        id: Optional[int] = None,
        *,
        tag: Optional[str] = None,
        max_bookmark_id: Optional[int] = None,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
    ) -> UserBookmarksIllustsResult:
        if id is None:
            id = self._client.account.id
        data = (
            await self._client.get(
                V1_API / "user/bookmarks/illust",
                params={
                    "user_id": id,
                    "filter": filter,
                    "tag": tag,
                    "max_bookmark_id": max_bookmark_id,
                    "restrict": "public",
                },
            )
        ).json()
        with set_pixiv_client(self._client):
            return UserBookmarksIllustsResult.parse_obj(data)

    async def novels(
        self,
        id: Optional[int] = None,
        *,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
        offset: int = None,
    ) -> UserNovelsResult:
        id = self._client.account.id if id is None else id
        data = (
            await self._client.get(
                V1_API / "user/novels",
                params={"user_id": id, "filter": filter, "offset": offset},
            )
        ).json()
        with set_pixiv_client(self._client):
            return UserNovelsResult.parse_obj(data)

    async def related(
        self,
        id: Optional[int] = None,
        *,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
        offset: int = None,
    ) -> UserRelatedResult:
        if id is None:
            id = self._client.account.id
        data = (
            await self._client.get(
                V1_API / "user/related",
                params={"seed_user_id": id, "filter": filter, "offset": offset},
            )
        ).json()
        with set_pixiv_client(self._client):
            return UserRelatedResult.parse_obj(data)



from datetime import date
from io import BytesIO
from pathlib import Path
from typing import (
    Dict,
    Iterator,
    List,
    Optional,
    Union,
)
from zipfile import ZipFile

from typing_extensions import Literal

from async_pixiv.client._section._base import (
    IllustType,
    SearchDuration,
    SearchFilter,
    SearchShort,
    SearchTarget,
    V1_API,
    V2_API,
    _Section,
)
from async_pixiv.error import ArtWorkTypeError
from async_pixiv.model.illust import IllustType
from async_pixiv.model.result import (
    IllustCommentResult,
    IllustDetailResult,
    IllustNewResult,
    IllustRelatedResult,
    IllustSearchResult,
    RecommendedResult,
    UgoiraMetadataResult,
)
from async_pixiv.utils.context import set_pixiv_client

__all__ = ("ILLUST",)


# noinspection PyShadowingBuiltins
class ILLUST(_Section):
    async def follow(self, *, offset: Optional[int] = None) -> IllustSearchResult:
        data = (
            await self._client.get(
                V2_API / "illust/follow",
                params={"restrict": "public", "offset": offset},
            )
        ).json()
        with set_pixiv_client(self._client):
            return IllustSearchResult.parse_obj(data)

    async def search(
        self,
        word: str,
        *,
        sort: Union[
            Literal["date_desc", "date_asc", "popular_desc", "popular_asc"], SearchShort
        ] = SearchShort.date_desc,
        target: Union[
            SearchTarget,
            Literal[
                "partial_match_for_tags", "keyword", "exact_match_for_tags", "text"
            ],
        ] = SearchTarget.partial,
        duration: Optional[
            Union[
                Literal[
                    "within_last_day",
                    "within_last_week",
                    "within_last_month",
                    "within_last_year",
                ],
                SearchDuration,
            ]
        ] = None,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
        offset: Optional[int] = None,
        min_bookmarks: Optional[int] = None,
        max_bookmarks: Optional[int] = None,
        start_date: Optional[date] = None,
        end_date: Optional[date] = None,
        **kwargs,
    ) -> IllustSearchResult:
        if start_date is not None:
            start_date = start_date.strftime("%Y-%m-%d")
        if end_date is not None:
            end_date = end_date.strftime("%Y-%m-%d")
        data = await super(ILLUST, self).search(
            word=word,
            sort=sort,
            duration=duration,
            filter=filter,
            offset=offset,
            bookmark_num_min=min_bookmarks,
            bookmark_num_max=max_bookmarks,
            target=target,
            start_date=start_date,
            end_date=end_date,
        )
        with set_pixiv_client(self._client):
            return IllustSearchResult.parse_obj(data)

    async def detail(
        self,
        id: int,
        *,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
    ) -> IllustDetailResult:
        data = await super(ILLUST, self).detail(id, filter=filter)
        with set_pixiv_client(self._client):
            return IllustDetailResult.parse_obj(data)

    async def comments(
        self, id: int, *, offset: Optional[int] = None
    ) -> IllustCommentResult:
        data = (
            await self._client.get(
                V1_API / "illust/comments", params={"illust_id": id, "offset": offset}
            )
        ).json()
        with set_pixiv_client(self._client):
            return IllustCommentResult.parse_obj(data)

    async def related(
        self,
        id: int,
        *,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
        offset: Optional[int] = None,
        seed_ids: Optional[Union[List[int], int]] = None,
    ) -> IllustRelatedResult:
        data = (
            await self._client.get(
                V2_API / "illust/related",
                params={
                    "illust_id": id,
                    "offset": offset,
                    "filter": filter,
                    "seed_illust_ids": (
                        [seed_ids] if seed_ids is not list else seed_ids
                    ),
                },
            )
        ).json()
        with set_pixiv_client(self._client):
            return IllustRelatedResult.parse_obj(data)

    async def recommended(
        self,
        *,
        with_auth: bool = True,
        type: Union[
            Literal["illust", "manga", "novel", "ugoira"], IllustType
        ] = IllustType.illust,
        include_ranking_label: bool = True,
        include_ranking_illusts: Optional[bool] = None,
        offset: Optional[int] = None,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
        max_bookmark_id_for_recommend: Optional[int] = None,
        min_bookmark_id_for_recent_illust: Optional[int] = None,
        bookmark_illust_ids: Optional[Union[List[int], int]] = None,
        include_privacy_policy: Optional[Union[str, List[Union[int, str]]]] = None,
        viewed: Optional[List[int]] = None,
    ) -> RecommendedResult:
        # noinspection SpellCheckingInspection
        data = (
            await self._client.get(
                V1_API
                / ("illust/recommended" if with_auth else "illust/recommended-nologin"),
                params={
                    "content_type": type,
                    "include_ranking_label": include_ranking_label,
                    "include_ranking_illusts": include_ranking_illusts,
                    "offset": offset,
                    "filter": filter,
                    "viewed": viewed,
                    "max_bookmark_id_for_recommend": max_bookmark_id_for_recommend,
                    "min_bookmark_id_for_recent_illust": min_bookmark_id_for_recent_illust,
                    "bookmark_illust_ids": bookmark_illust_ids,
                    "include_privacy_policy": include_privacy_policy,
                },
            )
        ).json()
        with set_pixiv_client(self._client):
            return RecommendedResult.parse_obj(data)

    async def new_illusts(
        self,
        type: Literal["illust", "manga"] = "illust",
        *,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
        max_illust_id: Optional[int] = None,
    ) -> IllustNewResult:
        data = (
            await self._client.get(
                V1_API / "illust/new",
                params={
                    "content_type": type,
                    "max_illust_id": max_illust_id,
                    "filter": filter,
                },
            )
        ).json()
        with set_pixiv_client(self._client):
            return IllustNewResult.parse_obj(data)

    async def ugoira_metadata(self, id: int) -> UgoiraMetadataResult:
        data = (
            await self._client.get(V1_API / "ugoira/metadata", params={"illust_id": id})
        ).json()
        with set_pixiv_client(self._client):
            return UgoiraMetadataResult.parse_obj(data)

    async def download(
        self,
        id: int,
        *,
        full: bool = False,
        filter: Optional[
            Union[Literal["for_android", "for_ios"], SearchFilter]
        ] = SearchFilter.ios,
        output: Optional[Union[str, Path, BytesIO]] = None,
    ) -> List[bytes]:
        artwork = (await self.detail(id, filter=filter)).illust
        if artwork.type == IllustType.ugoira:
            raise ArtWorkTypeError(
                "If you want to download a moving image, "
                'please use this method: "download_ugoira"'
            )
        if not full or not artwork.meta_pages:
            return [
                await self._client.download(
                    str(
                        artwork.meta_single_page.original
                        or artwork.image_urls.original
                        or artwork.image_urls.large
                    ),
                    output=output,
                )
            ]
        else:
            result: List[bytes] = []
            for meta_page in artwork.meta_pages:
                result.append(
                    await self._client.download(
                        str(
                            meta_page.image_urls.original or meta_page.image_urls.large
                        ),
                        output=output,
                    )
                )
            return result

    async def download_ugoira(
        self,
        id: int,
        *,
        type: Literal["zip", "jpg", "all", "iter"] = "zip",
    ) -> Optional[
        Union[
            ZipFile,
            List[bytes],
            Dict[str, Union[ZipFile, List[bytes]]],
            Iterator[bytes],
        ]
    ]:
        metadata = (await self.ugoira_metadata(id)).metadata
        zip_url = metadata.zip_url
        data = await self._client.download(
            zip_url.original or zip_url.large or zip_url.medium,
        )
        if data is None:
            return None
        zip_file = ZipFile(BytesIO(data))
        if type == "zip":
            return zip_file
        frames = []
        for frame_info in metadata.frames:
            with zip_file.open(frame_info.file) as f:
                frames.append(f.read())
        if type == "jpg":
            return frames
        elif type == "iter":
            return iter(frames)
        else:
            return {"zip": zip_file, "frames": frames}



from typing import Any, Optional, TYPE_CHECKING, TypeVar

from pydantic import (
    BaseConfig as PydanticBaseConfig,
    BaseModel as PydanticBaseModel,
    PrivateAttr,
    validator,
)

from async_pixiv.utils.context import PixivClientContext

if TYPE_CHECKING:
    from async_pixiv.client import PixivClient

    Model = TypeVar("Model", bound="BaseModel")

__all__ = ["PixivModel", "PixivModelConfig", "null_dict_validator"]

T = TypeVar("T")


class PixivModelConfig(PydanticBaseConfig):
    try:
        import ujson as json
    except ImportError:
        import json
    json_dumps = json.dumps
    json_loads = json.loads
    validate_all = True
    validate_assignment = True


class PixivModel(PydanticBaseModel):
    class Config(PixivModelConfig):
        pass

    _pixiv_client: Optional["PixivClient"] = PrivateAttr(None)

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._pixiv_client = PixivClientContext.get()

    def __new__(cls, *args, **kwargs):
        cls.update_forward_refs()
        return super(PixivModel, cls).__new__(cls)

    def __str__(self) -> str:
        if hasattr(self, "id"):
            return f'<{self.__class__.__name__} id="{self.id}">'
        else:
            return f"<{self.__class__.__name__}>"

    def __hash__(self) -> int:
        if hasattr(self, "id"):
            return hash((self.__class__.__name__, self.id))
        return super(PixivModel, self).__hash__()


def null_dict_validator(*fields: str) -> classmethod:
    def v(cls, value: T) -> Optional[T]:
        if value == {}:
            return None
        else:
            return value

    return validator(*fields, pre=True, always=True, allow_reuse=True)(v)



from async_pixiv.client._section._base import SectionType
from async_pixiv.client._section._illust import ILLUST
from async_pixiv.client._section._novel import NOVEL
from async_pixiv.client._section._user import USER



from contextlib import asynccontextmanager, contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from async_pixiv.client import PixivClient

__all__ = (
    "no_warning",
    "do_nothing",
    "async_do_nothing",
    "PixivClientContext",
    "set_pixiv_client",
)

PixivClientContext: ContextVar["PixivClient"] = ContextVar("PixivClientContext")


@contextmanager
def do_nothing() -> None:
    try:
        yield
    finally:
        ...


@asynccontextmanager
def async_do_nothing() -> None:
    try:
        yield
    finally:
        ...


@contextmanager
def no_warning() -> None:
    import warnings
    from copy import deepcopy

    filters = deepcopy(warnings.filters)
    try:
        warnings.filterwarnings("ignore")
        yield
    finally:
        warnings.resetwarnings()
        for f in filters:
            # noinspection PyUnresolvedReferences
            warnings.filters.append(f)


@contextmanager
def set_pixiv_client(client: "PixivClient") -> "PixivClient":
    token = PixivClientContext.set(client)
    try:
        yield client
    finally:
        PixivClientContext.reset(token)



from enum import IntEnum
from typing import Optional, TYPE_CHECKING

from pydantic import HttpUrl
from yarl import URL

from async_pixiv.model._base import PixivModel, PixivModelConfig

if TYPE_CHECKING:
    from async_pixiv import PixivClient
    from async_pixiv.model.result import NovelSeriesResult

__all__ = ["Series", "ImageUrl", "Tag", "TagTranslation", "AIType"]


class Series(PixivModel):
    id: int
    title: str

    @property
    def link(self) -> URL:
        return URL(f"https://www.pixiv.net/novel/series/{self.id}")

    async def detail(
        self, client: Optional["PixivClient"] = None
    ) -> Optional["NovelSeriesResult"]:
        if client is None:
            from async_pixiv.client import PixivClient

            client = PixivClient.get_client()
        return await client.NOVEL.series(self.id)


class ImageUrl(PixivModel):
    square_medium: Optional[HttpUrl]
    medium: Optional[HttpUrl]
    large: Optional[HttpUrl]
    original: Optional[HttpUrl]

    @property
    def link(self) -> HttpUrl:
        return self.original or self.large or self.medium or self.square_medium


class TagTranslation(PixivModel):
    class Config(PixivModelConfig):
        extra = "allow"

    zh: Optional[str]
    en: Optional[str]
    jp: Optional[str]


class Tag(PixivModel):
    name: str
    translated_name: Optional[str]
    added_by_uploaded_user: Optional[bool]

    def __init__(
        self,
        name: str,
        translated_name: Optional[str] = None,
        added_by_uploaded_user: Optional[bool] = None,
    ) -> None:
        super().__init__(
            name=name,
            translated_name=translated_name,
            added_by_uploaded_user=added_by_uploaded_user,
        )


class AIType(IntEnum):
    NONE = 0
    """没有使用AI"""

    HALF = 1
    """使用了AI进行辅助"""

    FULL = 2
    """使用AI生成"""



from enum import Enum as BaseEnum

__all__ = ("Enum", "SearchTarget", "SearchShort", "SearchDuration", "SearchFilter")


class Enum(BaseEnum):
    def __str__(self) -> str:
        # noinspection PyTypeChecker
        return self.value


class SearchTarget(Enum):
    partial = "partial_match_for_tags"  # 标签部分一致
    full = "exact_match_for_tags"  # 标签完全一致
    text = "text"  # 正文
    keyword = "keyword"  # 关键词


class SearchShort(Enum):
    date_desc = "date_desc"
    date_asc = "date_asc"
    popular_desc = "popular_desc"
    popular_asc = "popular_asc"


class SearchDuration(Enum):
    day = "within_last_day"
    week = "within_last_week"
    month = "within_last_month"
    year = "within_last_year"


class SearchFilter(Enum):
    android = "for_android"
    ios = "for_ios"
